/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.mergeall;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.BasicControllerMixin;
import org.objectweb.fractal.julia.control.binding.BasicBindingControllerMixin;
import org.objectweb.fractal.julia.control.binding.ContentBindingController;
import org.objectweb.fractal.julia.loader.BasicLoader;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.desc.SimpleMembraneDesc;
import org.objectweb.fractal.juliac.opt.FCSourceCodeGenerator;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.spoon.MixinClassGenerator;
import org.objectweb.fractal.juliac.spoon.SpoonHelper;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeWriter;

import spoon.reflect.Factory;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.factory.ClassFactory;
import spoon.reflect.reference.CtTypeReference;

/**
 * This class generates the source code associated to Fractal components.
 * Two classes are generated: the membrane implementation and the
 * initializer implementation.
 * 
 * The membrane implementation is generated by merging the content, the
 * interceptors and the controllers.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCMergeAllSourceCodeGenerator
extends FCSourceCodeGenerator<List<ControllerDesc>> {
    
    @Override
    protected void postInitJuliaLoaderModule() throws IOException {
        jloader.loadConfigFile(BasicLoader.DEFAULT_CONFIGURATION);
        super.postInitJuliaLoaderModule();
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Generate the implementation of the membrane associated to a Fractal
     * component.
     * 
     * @param ct                the component type
     * @param ctrldesc          the component controller descriptor
     * @param contentClassName  the name of the component content class
     * @param context           contextual information
     * @return                  data about the membrane implementation
     */
    public MembraneDesc<List<ControllerDesc>> generateMembraneImpl(
        ComponentType ct, String ctrldesc, String contentClassName )
    throws IOException {
        
        /*
         * Generate business and control interface implementations.
         */
        InterfaceType[] its = ct.getFcInterfaceTypes();
        generateInterfaceImpl(its,ctrldesc);
        InterfaceType[] membraneits = mloader.getMembraneType(ct,ctrldesc);
        generateInterfaceImpl(membraneits,ctrldesc);
        
        /*
         * Prepare a membrane descriptor.
         */
        MembraneDesc<List<ControllerDesc>> md =
            new SimpleMembraneDesc(jc,null,ctrldesc,membraneits,null);

        /*
         * Retrieve the list of interceptor and control layers.
         * Interceptor layers are the last ones. Last layers are executed first.
         */
        List<CtClass<?>> layers = new ArrayList<CtClass<?>>();
        List<CtClass<?>> interceptorLayers = getInterceptorLayers(ct,md);
        List<CtClass<?>> controlLayers = getControlLayers(ct,ctrldesc);
        layers.addAll(controlLayers);
        layers.addAll(interceptorLayers);
        
        /*
         * Layers specific to components with a content.
         */
        if( contentClassName != null ) {
            
            // Perform checks on the content class
            checkContentClass( contentClassName );
            
            List<CtClass<?>> contentCallLayers = getContentCallLayers(ct);
            layers.addAll(0,contentCallLayers);            

            /*
             * Membranes such as primitives, assume that the content provides an
             * implementation of the BindingController interface. If this is the
             * case, add the layer which will call this implementation.
             */
            CtClass<?> contentBCLayer =
                getContentBindingControllerLayer(ct,layers);
            if( contentBCLayer != null ) {
                layers.add(contentBCLayer);
            }
        }
        
        /*
         * Generate and register the class by merging the content, controllers
         * and interceptors. Avoid duplicate generation.
         */
        InitializerClassGenerator<List<ControllerDesc>> icg =
            getInitializerClassGenerator();
        icg.init(jc,this,null,null,null);
        String membraneClassName =
            icg.getMembraneClassName(ctrldesc,contentClassName);        

        if( ! jc.hasBeenGenerated(membraneClassName) ) {
            
            JuliacConfig jconf = jc.getJuliacConfig();
            SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
            Factory factory = spoon.getSpoonFactory();
            MixinClassGenerator spmcg = new MixinClassGenerator(factory);

            CtClass<?>[] ctLayers =
                layers.toArray(new CtClass<?>[layers.size()]);
            CtClass<?> generated = spmcg.generate(membraneClassName,ctLayers);
            
            /*
             * Set the superclass to be the content class.
             */
            if( contentClassName != null ) {
                ClassFactory cf = factory.Class();
                CtTypeReference<?> ref = cf.createReference(contentClassName);
                generated.setSuperclass(ref);
            }
            
            /*
             * Dump the generated class.
             */
            spoon.generateSourceCode(generated);        
        }
        
        /*
         * Prepare a membrane descriptor.
         */
        List<ControllerDesc> ctrlDescs = new ArrayList<ControllerDesc>();
        ctrlDescs.add(new ControllerDesc(membraneClassName,null,null));
        MembraneDesc<List<ControllerDesc>> newmd =
            new SimpleMembraneDesc(jc,null,ctrldesc,membraneits,ctrlDescs);
        newmd.setInterceptorClassGenerator(md.getInterceptorClassGenerator());

        /*
         * Set the interceptor source code generator for latter use in the
         * generation of the initializer class for generating interceptors on
         * client interfaces (interceptors on server interfaces are merged).
         */
        ProxyClassGeneratorItf itbcg =
        	getInterceptorClassGenerator(ct,ctrldesc);
        newmd.setInterceptorClassGenerator(itbcg);
        
        return newmd;
    }

    /**
     * Return the source code generator for the initializer class associated
     * with this component source code generator.
     */
    public InitializerClassGenerator<List<ControllerDesc>>
    getInitializerClassGenerator() {
        return new InitializerMergeAllClassGenerator();
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation of methods defined in FCSourceCodeGenerator
    // -----------------------------------------------------------------------
    
    /**
     * Return the source code generator for interceptors.
     */
    @Override
    protected ProxyClassGeneratorItf getInterceptorClassGenerator(
		ComponentType ct, String ctrldesc,
		InterceptorSourceCodeGeneratorItf[] iscgs ) {
        
    	InterceptorClassGenerator icg = new InnerInterceptorClassGenerator();
    	icg.setJuliac(jc);
    	icg.init(iscgs);
    	return icg;
    }
    
    /**
     * Perform checks on the content class.
     */
    protected void checkContentClass( String contentClassName )
    throws IOException {
        
        /*
         * Check that the binding control methods in the content class declare
         * in their throws clause NoSuchInterfaceException,
         * IllegalBindingException and IllegalLifeCycleException. In the merge
         * all mode, the membrane class extends the content class, the Java
         * compiler complains if the content class does not throw these
         * exceptions.
         * 
         * If the content class does not throw these exceptions, report a
         * warning and generate runtime exceptions.
         */

        UnifiedClass bcuc = jc.create(BindingController.class.getName());
        UnifiedClass cl = jc.create(contentClassName);
        if( ! bcuc.isAssignableFrom(cl) ) {
            return;
        }
        
        UnifiedMethod mbind=null, mlookup=null, munbind=null;
        try {
            mbind =
                cl.getMethod(
                    "bindFc", new Class<?>[]{String.class,Object.class} );
            mlookup = cl.getMethod("lookupFc", new Class<?>[]{String.class});
            munbind = cl.getMethod("unbindFc", new Class<?>[]{String.class});
        } 
        catch (NoSuchMethodException e) {
            // As cl implements BindingController, these methods should be found
            throw new JuliacRuntimeException(e);
        }
        
        UnifiedMethod[] meths = new UnifiedMethod[]{mbind,mlookup,munbind};
        checkMethodsThrowsException(
            contentClassName, meths, NoSuchInterfaceException.class );
        
        meths = new UnifiedMethod[]{mbind,munbind};
        checkMethodsThrowsException(
            contentClassName, meths, IllegalBindingException.class );
        checkMethodsThrowsException(
            contentClassName, meths, IllegalLifeCycleException.class );
    }
    
    /**
     * Check that the specified methods declare the specified exception.
     * 
     * If this is not the case, generate the source code of a {@link
     * RuntimeException} for the specified exception.
     */
    private void checkMethodsThrowsException(
        String contentClassName, UnifiedMethod[] meths,
        Class<? extends Throwable> e )
    throws IOException {

        String ms = "";
        for (UnifiedMethod meth : meths) {
            boolean t = checkThrowsException(meth,e);
            if(!t) ms+= meth.getName()+" ";
        }
        
        if( ms.length() > 0 ) {
            String msg =
                "Method(s) "+ms+"in class "+contentClassName+ " should throw "+
                e.getName()+
                ". This exception will be weaken to a RuntimeException.";
            JuliacConfig jconf = jc.getJuliacConfig();
            Logger logger = jconf.getLogger();
            logger.warning(msg);

            String targetClassName = e.getName();
            ClassGeneratorItf scg =
                new RuntimeExceptionClassGenerator(targetClassName);
            jc.generateSourceCodeOverride(scg);
        }
    }
    
    /**
     * Return <code>true</code> if the specified method declares the specified
     * exception in its throws clause.
     */
    private boolean checkThrowsException(
        UnifiedMethod m, Class<? extends Throwable> e ) {
        
        UnifiedClass euc = jc.create(e.getName());
        UnifiedClass[] exceptions = m.getExceptionTypes();
        for (UnifiedClass exception : exceptions) {
            if( euc.isAssignableFrom(exception) ) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Return the list of mixin layers which call the content class (super
     * call).
     */
    private List<CtClass<?>> getContentCallLayers( ComponentType ct ) {
        
    	JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        
        ContentCallCtClassGenerator ccscg =
            new ContentCallCtClassGenerator(factory);
        
        List<CtClass<?>> contentCallLayers = new ArrayList<CtClass<?>>();
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (InterfaceType it : its) {
            
            if( it.isFcClientItf() ) {
                continue;   // Skip client interface types
            }

            CtClass<?> contentCallLayer = ccscg.generateCtClass(it,jc);
            contentCallLayers.add(contentCallLayer);            
        }
        
        return contentCallLayers;
    }
    
    /**
     * Return the list of mixin layers which implement the interceptor code for
     * the specified component type and the specified membrane descriptor.
     */
    private List<CtClass<?>> getInterceptorLayers(
        ComponentType ct, MembraneDesc<?> membraneDesc )
    throws IOException {
        
        List<CtClass<?>> interceptorLayers = new ArrayList<CtClass<?>>();

        String ctrldesc = membraneDesc.getDescriptor();
        boolean mergeable = true;
        InterceptorSourceCodeGeneratorItf[] scgs =
        	getInterceptorSourceCodeGenerators(ct,ctrldesc,mergeable);
        if( scgs == null ) {
            // No interceptor
            return interceptorLayers;
        }
        
    	JuliacConfig jconf = jc.getJuliacConfig();
    	SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        
        for (int i = 0; i < scgs.length; i++) {
            scgs[i].setMembraneDesc(membraneDesc);
            
            /*
             * The set of interface signatures for which an interception layer
             * has already been generated.
             */ 
            Set<String> signatures = new HashSet<String>();
            
            InterfaceType[] its = ct.getFcInterfaceTypes();
            for (InterfaceType it : its) {
                
                // Skip client and attribute-controller            
                if( it.isFcClientItf() ||
                    it.getFcItfName().equals("attribute-controller") ) {
                    continue;
                }
                
                // Skip if a layer has already been generated for this signature
                String signature = it.getFcItfSignature();
                if( signatures.contains(signature) ) {
                    continue;
                }

                /*
                 * Create the interceptor class generator.
                 */
                scgs[i].setInterfaceType(it);
                InterceptorClassGenerator icg =
                    new InterceptorClassGenerator(
                        jc,new InterceptorSourceCodeGeneratorItf[]{scgs[i]},
                        it, membraneDesc, mergeable );
                                
                /*
                 * Generate the code of the interceptor.
                 */
                StringWriter sw = new StringWriter();
                PrintWriter pw = new PrintWriter(sw);
                FileSourceCodeVisitor fw = new FileSourceCodeWriter(pw);
                icg.generate(fw);
                pw.close();
                sw.close();
                String interceptorSnippet = sw.toString();
                
                /*
                 * Populate the already existing Spoon model with the
                 * interceptor class.
                 */
                List<SourceFile> sfs = new ArrayList<SourceFile>();
                String targetClassname = icg.getTargetClassName();
                SourceFile sf =
                    SourceFile.create(targetClassname,interceptorSnippet);
                sfs.add(sf);
                spoon.build(sfs);
                
                /*
                 * Retrieve the CtClass corresponding to the interceptor class
                 * and add it in the list of interception layers.
                 */
                CtClass<?> interceptorLayer =
                    factory.Class().get(targetClassname);
                if( interceptorLayer == null ) {
                    String msg = "No such CtClass: "+targetClassname;
                    throw new IllegalArgumentException(msg);
                }

                interceptorLayers.add(interceptorLayer);
                signatures.add(signature);
            }
        }        
        
        return interceptorLayers;
    }
    
    /**
     * Return the layers associated with the specified control descriptor.
     */
    private List<CtClass<?>> getControlLayers(
        ComponentType ct, String ctrlDesc ) {
        
        List<String> layers = new UniqueArrayList<String>();
        
        List<ControllerDesc> cdescs = mloader.getCtrlImplLayers(ct,ctrlDesc);
        for (ControllerDesc cdesc : cdescs) {
			List<String> cdescLayers = cdesc.getLayerImplNames();
			if( cdescLayers == null ) {
                /*
                 * Special case where the name of the controller implementation
                 * is directly specified (not generated with
                 * MixinClassGenerator).
                 */
				String cdescImplName = cdesc.getImplName();
				layers.add(cdescImplName);
			}
			else {
				layers.addAll(cdescLayers);
			}
		}

        String bcmName = BasicControllerMixin.class.getName();
        if( layers.contains(bcmName) ) {
            /*
             * Julia assumes that BasicControllerMixin is always the first layer
             * for controller. This layer does not propagate calls to
             * initFcController. Julia reports an error if one tries to modify
             * this (not super method). The policy implemented by
             * MembraneMixinLayersDef for managing duplicate layers may lead to
             * cases where BasicControllerMixin is not the first layer if it
             * appears several times (only the last occurrence is kept.) In
             * order to deal with such a situation, reintroduce
             * BasicControllerMixin as the first layer.
             */
            layers.add(0,bcmName);
        }
        
    	JuliacConfig jconf = jc.getJuliacConfig();
    	SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        List<CtClass<?>> ctLayers = SpoonHelper.toCtClass(factory,layers);
        
        return ctLayers;
    }
    
    /**
     * Control membranes such as primitive expect that the content class
     * provides an implementation for the {@link BindingController} interface.
     * When this is the case, a layer implementing the {@link
     * ContentBindingController} must be provided. This method returns this
     * layer or <code>null</code> if the layer is not needed.
     */
    protected CtClass<?> getContentBindingControllerLayer(
        ComponentType ct, List<CtClass<?>> layers ) {
        
        if( ! isAssignableFrom(layers,BindingController.class) ||
            layers.contains(BasicBindingControllerMixin.class) ||
            ! containsClientInterfaceType(ct) ) {
            
            return null;
        }
        
        /*
         * ContentBindingControllerSourceCodeGenerator generates a default
         * implementation of the ContentBindingController interface. This
         * implementation is expected by control membranes such as
         * primitives, which expect that their content class will provides
         * an implementation of the BindingController interface (if the
         * component defines a client interface).
         */

    	JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        ContentBindingControllerCtClassGenerator cbcscg =
            new ContentBindingControllerCtClassGenerator(factory);
        
        InterfaceType it =
            new BasicInterfaceType(
                "cbci",ContentBindingController.class.getName(),
                false,false,false);
        CtClass<?> generated = cbcscg.generateCtClass(it,jc);
        return generated;
    }
    
    /**
     * Return <code>true</code> if the specified component type defines a client
     * interface.
     */
    private static boolean containsClientInterfaceType( ComponentType ct ) {
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (InterfaceType it : its) {
            if( it.isFcClientItf() ) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Return <code>true</code> if the specified interface is assignable from
     * one of the classes of the specified list.
     */
    private boolean isAssignableFrom( List<CtClass<?>> layers, Class<?> itf ) {
        
    	JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        CtTypeReference<?> itftr = factory.Type().createReference(itf);
        
        for (CtClass<?> layer : layers) {
            CtTypeReference<?> layertr = layer.getReference();
            if( itftr.isAssignableFrom(layertr) ) {
                return true;
            }
        }
        return false;
    }
}
